import os
import csv
import pandas as pd
import numpy as np

# A magic number the _could be read/inferrred from the data, but let's
# be real: It's not changing.
NUM_CUB_CLASSES = 200
NUM_CUB_ATTRIBUTES = 312

def load_attribute_schema(dataset_path='',clean=True):
# Return dict with key as the attribute id (int type) and value a
# 2-tuple (attribute,value) where attribute is the description of the
# attribute and value the specific value it takes, ie.,
# ('has_bill_shape','dagger') clean indicates whether to alter the
# values for use in textual descriptions according to pre-codified
# rules.
    
    attributes = {}
    with open(os.path.join(dataset_path,'attributes.txt')) as fd:
        # Each line has the form:
        # 12 has_wing_color::iridescent
        for attr_line in fd:
            [attr_id,attr_name_val] = attr_line.strip().split(' ')
            attributes[int(attr_id)] = attr_name_val.split('::')

    if clean: # Hacky list of alterations to make for captioning
        # has_bill_shape
        attributes[1][1] = 'curved' # curved_(up_or_down)

        # has_tail_shape
        for a in range(74,80): # 74:79
            attributes[a][1] = attributes[a][1].removesuffix('_tail')

        # has_bill_length
        attributes[150][1] = '' # about_the_same_as_head
        attributes[151][1] = 'long' # longer_than_head
        attributes[152][1] = 'short' # shorter_than_head

        # has_wing_shape
        for a in range(213,218): # 213:217
            attributes[a][1] = attributes[a][1].removesuffix('-wings')
            
    return attributes
            
def load_class_attributes(dataset_path=''):
# Return Pandas DataFrame with indices being the CUB class id and the
# columns being attribute IDs. Ebtry value is the percentage of the
# time (0-100) the class is imputed with that attribute.

    attr_arr = np.loadtxt(
        os.path.join(
            dataset_path,'attributes/class_attribute_labels_continuous.txt'))

    return pd.DataFrame(attr_arr,
                        index=list(range(1,NUM_CUB_CLASSES+1)),
                        columns=list(range(1,NUM_CUB_ATTRIBUTES+1)) )

    
                 
class CUB():

    def __init__(self,dataset_path='', map_file=None, attr_thresh=50.0):

        self.attr_thresh = attr_thresh
        
        # Load in the attributes schema
        self.attr_schema = load_attribute_schema(dataset_path)
        
        # Load the attribute matrix
        self.attributes = load_class_attributes(dataset_path)

        # Invert to access e.g., 'has_eye_color' --> [136, 137, ...]
        self.invert_attribute_index()

        # Load (species) conversion from NABird ID to CUB ID
        if map_file is not None:
            with open(map_file) as fd:
                self.nab_map = { int(row['nabird_id']):int(row['cub_id'])
                                 for row in csv.DictReader(fd) }


    def get_id(self,nabird_id):
        # Produce the CUB id (int)
        if nabird_id in self.nab_map:
            return self.nab_map[nabird_id]
        else:
            return None
    
    def invert_attribute_index(self):
        # Produce/store a dictionary mapping attribute name to a list
        # of integers (cub_attr_id values) for that name. I.e.,
        # {'has_bill_shape': [1,2,3,4,5,6,7,8,9], ... }

        inv_index = {}

        # Iterate over all attribute ids, re-indexing by names (building lists)
        for (attr_id,attr_name_val) in self.attr_schema.items():

            [attr_name,attr_val] = attr_name_val
            
            if attr_name in inv_index: # key already added?
                inv_index[attr_name].append(attr_id) # append id to list
            else:
                inv_index[attr_name] = [attr_id] # initialize list with id

        # Ensure a canonical ordering by sorting names in list values
        for attr_names in inv_index.values():
            attr_names.sort() 
        
        self.attr_index = inv_index
                
        
    def attribute_values(self,bird_id,attr_name):
        # Get the list of an attribute's descriptions for a given bird species

        attr_ids = self.attr_index[attr_name] # list of indices for attribute

        # Pandas magic
        bird_attr = self.attributes.loc[bird_id]  # row for given bird
        #print('row',bird_attr)
        bird_attr = bird_attr[attr_ids]           # cols for given attribute name
        #print('cols',bird_attr)
        bird_attr = bird_attr > self.attr_thresh  # find specific attr values
        #print('matches',bird_attr)
        bird_attr = bird_attr.index[bird_attr]    # extract matching values
        #print('indices',bird_attr)

        # extract names of matching values
        names = [self.attr_schema[attr_id][1] for attr_id in bird_attr]

        return names



# Note: Each subclass's attr field is given in canonical/hierarchical
# adjective order. See
# https://dictionary.cambridge.org/us/grammar/british-grammar/adjectives-order

# At the 50% threshold, only color attributes have multiple
# matches. Since there are no guarantees this exclusion of other
# attributes would hold at lower thresholds, the code applies the
# possibility to all attributes. It inserts "and" at the conclusion of
# such a multi-match sequence, even if there are no commas to
# distinguish/group as a sequence. Most of the time this will work
# because color is last in the ordering (except for left/right
# directions).

class Part():
    cub = None # Singleton object, shared by all derivative parts

    def describe(self,nab_id):
        # Get all the adjectives

        descr= '' 
         
        cub_id = self.cub.get_id(nab_id)

        if cub_id is None: # Without a matching CUB id, there are no part  
            return descr
        
        for name in self.attrs: # i.e., length, shape, color
            vals = self.cub.attribute_values(cub_id,name) # i.e., ['blue', 'red']
            if len(vals) == 0:
                continue # No matching attribute annotation found
            if len(vals) == 1:
                descr += ' ' + vals[0] # Append description of single value
            else:
                # Append descriptions of all values with an "and" before the last
                descr += ' ' + ' '.join(vals[0:-1])
                descr += ' and ' + vals[-1]
                
        return descr.strip() # Leading space


    def description(self,nab_id):
    # Get all the adjectives plus the noun
        descr = self.describe(nab_id)
        if len(descr)>0:
            return descr + ' ' + self.name
        else:
            return self.name


class OrientedPart(Part):
    def __init__(self,sagittal_dir):
        super().__init__()
        if not (sagittal_dir in ['left','right']):
            raise ValueError('sagittal_dir must be "left" or "right"')
        self.sagittal_dir = sagittal_dir

    def describe(self,nab_id):
        # Appends direction to description
        
        descr = super().describe(nab_id)
        if len(descr)>0: # suggestions on cleaner/idiomatic ways to do 
            descr += ' ' # conditional space insertion welcome
        descr += self.sagittal_dir
        return descr
    
    
class Bill(Part):
    attrs = ['has_bill_length','has_bill_shape','has_bill_color'] # 2,4,6
    name = 'bill'
    
class Crown(Part):
    attrs = ['has_crown_color']
    name = 'crown'
    
class Nape(Part):
    attrs = ['has_nape_color']
    name = 'nape'

class Eye(OrientedPart):
    attrs = ['has_eye_color']
    name = 'eye'
    

class Belly(Part):
    attrs = ['has_belly_pattern', 'has_belly_color'] # 3,6
    name = 'belly'
    
class Breast(Part):
    attrs = ['has_breast_pattern', 'has_breast_color'] # 3,6
    name = 'breast'
    
class Back(Part):
    attrs = ['has_back_pattern', 'has_back_color'] # 3,6
    name = 'back'
    
class Tail(Part):
    attrs = ['has_tail_pattern', 'has_tail_shape', 'has_upper_tail_color',
                 'has_under_tail_color'] # 3, 4, 6, 6
    name = 'tail'
    
class Wing(OrientedPart):
    attrs = ['has_wing_pattern','has_wing_shape','has_wing_color'] # 3, 4, 6
    name = 'wing'
    
        
class NABParts():
    # Parts listed in NABird part_id order-- [0,10] -- cf. NABirds/parts/parts.txt
    parts = [ Bill(),
              Crown(),
              Nape(),
              Eye('left'),
              Eye('right'),
              Belly(),
              Breast(),
              Back(),
              Tail(),
              Wing('left'),
              Wing('right') ]



#!/usr/bin/env python3

import os
import json

def extract(coco_file,map_file):
    
    with open(coco_file) as fd:
        coco_data = json.loads(fd.read())

    with open(map_file,'w') as fd:
        for entry in coco_data['images']:
            img_fname = entry['file_name']
            coco_id = entry['id']
            fhash = os.path.basename(img_fname).split('.')[0]
            nabird_id = '{}-{}-{}-{}-{}'.format(fhash[0:8],
                                                fhash[8:12],
                                                fhash[12:16],
                                                fhash[16:20],
                                                fhash[20:])

            fd.write(f'{coco_id} {nabird_id}\n')

if __name__ == '__main__':

    for split in ['train','val','test']:
        extract(f'../data/{split}.json',f'../data/{split}.map')



# Copied from the NABirds dataset; presumably written by GVH.
# Updated for python3 by Jerod
import os
import sys

#from matplotlib import pyplot as plt
import random
import math


def load_bounding_box_annotations(dataset_path=''):
  
  bboxes = {}
  
  with open(os.path.join(dataset_path, 'bounding_boxes.txt')) as f:
    for line in f:
      pieces = line.strip().split()
      image_id = pieces[0]
      bbox = [int(x) for x in pieces[1:]]
      bboxes[image_id] = bbox
  
  return bboxes

def load_part_annotations(dataset_path=''):
  
  parts = {}
  
  with open(os.path.join(dataset_path, 'parts/part_locs.txt')) as f:
    for line in f:
      pieces = line.strip().split()
      image_id = pieces[0]
      parts.setdefault(image_id, [0] * 11)
      part_id = int(pieces[1])
      parts[image_id][part_id] = [int(x) for x in pieces[2:]]

  return parts  
  
def load_part_names(dataset_path=''):
  
  names = {}

  with open(os.path.join(dataset_path, 'parts/parts.txt')) as f:
    for line in f:
      pieces = line.strip().split()
      part_id = int(pieces[0])
      names[part_id] = ' '.join(pieces[1:])
  
  return names  
    
def load_class_names(dataset_path=''):
  
  names = {}
  
  with open(os.path.join(dataset_path, 'classes.txt')) as f:
    for line in f:
      pieces = line.strip().split()
      class_id = pieces[0]
      names[class_id] = ' '.join(pieces[1:])
  
  return names

def load_image_labels(dataset_path=''):
  labels = {}
  
  with open(os.path.join(dataset_path, 'image_class_labels.txt')) as f:
    for line in f:
      pieces = line.strip().split()
      image_id = pieces[0]
      class_id = pieces[1]
      labels[image_id] = class_id
  
  return labels
        
def load_image_paths(dataset_path='', path_prefix=''):
  
  paths = {}
  
  with open(os.path.join(dataset_path, 'images.txt')) as f:
    for line in f:
      pieces = line.strip().split()
      image_id = pieces[0]
      path = os.path.join(path_prefix, pieces[1])
      paths[image_id] = path
  
  return paths

def load_image_sizes(dataset_path=''):
  
  sizes = {}
  
  with open(os.path.join(dataset_path, 'sizes.txt')) as f:
    for line in f:
      pieces = line.strip().split()
      image_id = pieces[0]
      width, height = [int(x) for x in pieces[1:]]
      sizes[image_id] = [width, height]
  
  return sizes

def load_hierarchy(dataset_path=''):
  
  parents = {}
  
  with open(os.path.join(dataset_path, 'hierarchy.txt')) as f:
    for line in f:
      pieces = line.strip().split()
      child_id, parent_id = pieces
      parents[child_id] = parent_id
  
  return parents

def load_photographers(dataset_path=''):
  
  photographers = {}
  with open(os.path.join(dataset_path, 'photographers.txt')) as f:
    for line in f:
      pieces = line.strip().split()
      image_id = pieces[0]
      photographers[image_id] = ' '.join(pieces[1:])
  
  return photographers

def load_split(dataset_path='',split_file=None):
  train_images = []
  test_images  = []
  val_images   = []
  
  if split_file == None:
    split_file = os.path.join(dataset_path, 'train_test_split.txt')
    
  with open(split_file) as f:
    for line in f:
      pieces = line.strip().split()
      image_id = pieces[0]
      split = int(pieces[1])
      if split==1:
        train_images.append(image_id)
      elif split==2:
        val_images.append(image_id)
      else:
        test_images.append(image_id)
        
  return train_images, test_images, val_images

class NABirds():

    def __init__(self,dataset_path='', split_file=None, path_prefix='images'):

        # Load in the image data
        self.image_paths = load_image_paths(dataset_path, path_prefix)
        self.image_sizes = load_image_sizes(dataset_path)
        self.image_bboxes =load_bounding_box_annotations(dataset_path)
        self.image_parts = load_part_annotations(dataset_path)
        self.image_class_labels = load_image_labels(dataset_path)

        # Load in the class data
        self.class_names = load_class_names(dataset_path)
        self.class_hierarchy = load_hierarchy(dataset_path)

        # Load in the part data
        self.part_names = load_part_names(dataset_path)

        # Load in the photographers
        self.photographers = load_photographers(dataset_path)

        # Load in the train / test split
        self.train_images, self.test_images, self.val_images = load_split(dataset_path,split_file)




    def image_identifiers(self):
        return list(self.image_paths.keys())

    def part_ids(self):
        return sorted(self.part_names.keys())


if __name__ == '__main__':
  
  if len(sys.argv) > 1:
    dataset_path = sys.argv[1]
  else:
    dataset_path = ''
  
  if len(sys.argv) > 2:
    image_path = sys.argv[2]
  else:
    image_path  = 'images'
  
  # Load in the image data
  # Assumes that the images have been extracted into a directory called "images"
  image_paths = load_image_paths(dataset_path, path_prefix=image_path)
  image_sizes = load_image_sizes(dataset_path)
  image_bboxes = load_bounding_box_annotations(dataset_path)
  image_parts = load_part_annotations(dataset_path)
  image_class_labels = load_image_labels(dataset_path)
  
  # Load in the class data
  class_names = load_class_names(dataset_path)
  class_hierarchy = load_hierarchy(dataset_path)
  
  # Load in the part data
  part_names = load_part_names(dataset_path)
  part_ids = part_names.keys()
  part_ids = sorted(part_ids)

  # Load in the photographers
  photographers = load_photographers(dataset_path)
  
  # Load in the train / test split
  train_images, test_images = load_train_test_split(dataset_path)
  
  # Visualize the images and their annotations
  image_identifiers = list(image_paths.keys())
  random.shuffle(image_identifiers) 
  for image_id in image_identifiers:
    
    image_path = image_paths[image_id]
    image = plt.imread(image_path)
    bbox = image_bboxes[image_id]
    parts = image_parts[image_id]
    class_label = image_class_labels[image_id]
    class_name = class_names[class_label]

    # Render the image
    plt.close(1)
    fig = plt.figure(1, figsize=(16, 12))
    plt.imshow(image)
    plt.title("Image ID: %s \n Class Label: %s  [ID: %s]" % (image_id, class_name, class_label))
    
    # Render the bounding box annotations
    bbox_x, bbox_y, bbox_width, bbox_height = bbox
    currentAxis = plt.gca()
    currentAxis.add_patch(plt.Rectangle((bbox_x , bbox_y), bbox_width, bbox_height, facecolor="green", alpha=0.3))
    
    # Render the part annotations
    if has_networkx:
      # Use networkx spring layout
      
      G = nx.Graph()
      part_data = []
      initial_positions = {}
      for part_id, part in enumerate(parts):
        x, y, v = part
        if v: 
          
          part_str = 'part_%d' % (part_id,)
          label_str = 'label_%d' % (part_id,)
          
          G.add_node(part_str)
          G.add_node(label_str)
          G.add_edge(part_str, label_str)
          
          part_data.append(part_str)
          
          initial_positions[part_str] = (x, y)
          initial_positions[label_str] = (x, y)
          
      positions = nx.spring_layout(G, dim=2, k=20.0, pos=initial_positions, fixed=part_data)
      
      for part_id, part in enumerate(parts):
        x, y, v = part
        if v: 
          
          plt.plot(x, y, 'o')
          
          label_str = 'label_%d' % (part_id,)
          label_position = positions[label_str]
          label_xy = (label_position[0] * bbox_width + bbox_x, label_position[1] * bbox_height + bbox_y)
          plt.annotate(
            part_names[part_id], 
            xy=(x, y),
            xycoords='data', 
            xytext= label_xy, 
            textcoords='data',
            ha='right', 
            va='bottom',
            bbox=dict(boxstyle= 'round,pad=0.5', 
                  fc='yellow', 
                  alpha=0.5),
            arrowprops=dict(arrowstyle='->', 
                    connectionstyle='arc3,rad=0')
          )
  
    else:
      # just a basic label annotation for the part locations
      offset = (2 * math.pi) / len(parts)
      for part_id, part in enumerate(parts):
        x, y, v = part
        if v: 
          # try to offset the part labels so that they don't overlap too badly
          dist = random.randint(35, 65)
          offset_x = dist * math.cos(offset * part_id)
          offset_y = dist * math.sin(offset * part_id)
        
          plt.plot(x, y, 'o')
          plt.annotate(
            part_names[part_id], 
            xy=(x, y), 
            xytext=(offset_x, offset_y), 
            textcoords='offset points', 
            ha='right', 
            va='bottom',
            bbox=dict(boxstyle= 'round,pad=0.5', 
                  fc='yellow', 
                  alpha=0.5),
            arrowprops=dict(arrowstyle='->', 
                    connectionstyle='arc3,rad=0')
          )
    
    plt.show(block=False)
    
    # Print some info about the image and the annotations:
    print("Image ID: {}".format(image_id))
    print("Image Path: {}".format(image_path))
    print("Image Dimensions: {} x {}".format(image_sizes[image_id][0], image_sizes[image_id][1]))
    print("Photographer: {}".format(photographers[image_id]))
    print("Split: {}".format("train" if image_id in train_images else "test"))
    print("Class Label Path:")
    
    label_path = []
    current_class_label = class_label
    while current_class_label in class_hierarchy:
      label_path.append((class_names[current_class_label], current_class_label))
      current_class_label = class_hierarchy[current_class_label]
    label_path.append((class_names[current_class_label], current_class_label)) # add the root node
    label_path.reverse()
    for depth, (class_name, class_id) in enumerate(label_path):
      line = "\t" * (depth + 1)
      print("{} |- {} \t[ID: {}]".format(line,class_name, class_id))
  
    print("Bounding Box Dimensions: {} x {}".format(bbox_width, bbox_height))
    print("{}\t{}".format('Part Name', 'Visibility'))
    for part_id in part_ids:
      x, y, v = parts[part_id]
      print("{}\t{}".format(part_names[part_id], 'Y' if v else 'N'))
    
    # Keep showing images until a button other than Enter is pressed
    ri = input("Press Enter to continue. Or press any key and then enter to quit...")
    if ri != '':
      break
    print
    print

  
  
  



#!/usr/bin/env python3

import os

from echo1_coco_builder.annotations_builder import CocoAnnotationsBuilder
from deps import getdeps # PYTHONPATH=$DATA_REPOSITORY/utils/python
from nabirds import NABirds
from cub import CUB
import parts

PBOXSZ = 20
part_id_base = 2
CAPTION_SEP = '.'

def get_caption(image_key,birds):
    """Generate caption string for the image_key and the map (keyed to
    category_id) of token (beg,end) position tuples for each section of the
    caption.

    """
    
    caption = 'bird'
    indices = { 1: (0,len(caption)) }
               
    for part_num in range(len(birds.image_parts[image_key])):
        nab_id = birds.image_class_labels[image_key] # bird species class id (str)
        [px,py,pv] = birds.image_parts[image_key][part_num]

        if pv==0: # Part not visible
            continue
        part_descr = parts.NABParts.parts[part_num].description(int(nab_id))
        caption += CAPTION_SEP + ' ' # Add separators ...
        # cf. https://github.com/microsoft/GLIP/blob/91f20921e691922167d8fa94fb55fa7a25ffe0ff/maskrcnn_benchmark/data/datasets/od_to_grounding.py#L46
        start_index = len(caption)
        caption += part_descr
        end_index = len(caption)
        indices[part_id_base+part_num] = (start_index,end_index)

    caption += '.'

    return caption, indices


def process_parts(part_id_base,birds,coco_data):
    # Process part categories    
    for class_id,class_name in birds.part_names.items():
        coco_data.add_category({
            "id"  : int(class_id) + part_id_base,
            "name": class_name # Species-generic class name
        })

def process_images(image_keys, birds, coco_data):
    image_id = 0 # COCO wants ints; no non-hacky way to associate NABird's hash :(

    image_ids = {}
    image_tokens = {}
    
    for image_key in image_keys:
        image_ids[image_key] = image_id
        image_width,image_height = birds.image_sizes[image_key]
        caption,indices = get_caption(image_key,birds)

        # Captions are generated in part_num sorted order, so let's make sure the
        # tokens positions are in the same order
        tokens = [[indices[part]] for part in sorted(indices.keys())]
        image_tokens[image_key] = indices
        coco_data.add_image({
            "id"       : image_id,
            "file_name": birds.image_paths[image_key],
            "width"    : image_width,
            "height"   : image_height,
            "caption"  : caption,
            "tokens"   : tokens
        })
        image_id = image_id + 1

    return image_ids,image_tokens

def process_bboxes(image_keys, image_ids, image_tokens, part_base, birds,
                   coco_data):

    annotation_id = 0
    for image_key in image_keys:
        nab_id = birds.image_class_labels[image_key] # bird species class id (str)
        # Add bird bbox
        bird_bbox = birds.image_bboxes[image_key] # [x,y,width,height]
        coco_data.add_annotation({
            "id"         : annotation_id,
            "image_id"   : image_ids[image_key],
            "category_id": 1,
            "bbox"       : bird_bbox,
            "iscrowd"    : 0,
            "area"       : bird_bbox[2]*bird_bbox[3], # w * h
            "tokens"     : [image_tokens[image_key][1]]
            })
        annotation_id = annotation_id + 1
        
        # Add visible part bboxes
        for part_num in range(len(birds.image_parts[image_key])):
            [px,py,pv] = birds.image_parts[image_key][part_num]

            if pv==0: # Part not visible
                continue

            part_bbox = [px-PBOXSZ, py-PBOXSZ, 2*PBOXSZ, 2*PBOXSZ]

            part_descr = parts.NABParts.parts[part_num].description(int(nab_id))
            category_id = part_id_base+part_num
            coco_data.add_annotation({
                "id"         : annotation_id,
                "image_id"   : image_ids[image_key],
                "category_id": category_id,
                "bbox"       : part_bbox,
                "iscrowd"    : 0,
                "area"       : part_bbox[2]*part_bbox[3], # w * h
                "name"       : part_descr,
                "tokens"     : [image_tokens[image_key][category_id]]
            })
            annotation_id = annotation_id + 1
        

def process_categories(output_path, birds):
    coco_data = CocoAnnotationsBuilder()
    coco_data.add_category({
        "id"  : 1,
        "name": "bird"
    })

    process_parts(part_id_base, birds, coco_data)
    with open(output_path,'w') as fd:
        fd.write(coco_data.get())

def process_split(output_path, image_keys, birds):
    
    coco_data = CocoAnnotationsBuilder()

    coco_data.add_info({
        "year"        : 2022,
        "version"     : "0.1",
        "description" : "Bird part points given bounding boxes (40px h/w)",
        "contributor" : "NABirds, Jerod Weinman",
        "date_created": "2022/07/15"
    })

    # bird + part categories
    coco_data.add_category({"id"  : 1,"name": "bird"})
    process_parts(part_id_base, birds, coco_data)

    image_ids,image_tokens = process_images(image_keys, birds, coco_data)

    process_bboxes(image_keys, image_ids, image_tokens, part_id_base,
                   birds, coco_data)

    with open(output_path,'w') as fd:
        fd.write(coco_data.get())
        
        
if __name__ == '__main__':
    
    deps = getdeps()
    birds = NABirds( deps['nab'],
                     os.path.join(deps['split'],'train_val_test_split.txt'),
                     'images')
    parts.Part.cub = CUB( deps['cub'],
                          os.path.join(deps['map'],'nabird_cub.csv'))
    process_split('../data/train.json', birds.train_images, birds)
    process_split('../data/val.json', birds.val_images, birds)
    process_split('../data/test.json', birds.test_images, birds)
    process_categories('../data/categories.json',birds)
